import json
import uuid
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Optional

from uipath import UiPath
from uipath.models import CreateAction, InvokeProcess, WaitAction, WaitJob

from .._runtime._contracts import (
    UiPathApiTrigger,
    UiPathErrorCategory,
    UiPathResumeTrigger,
    UiPathResumeTriggerType,
    UiPathRuntimeError,
    UiPathRuntimeStatus,
)
from .._utils._common import serialize_object
from ._escalation import Escalation


def _try_convert_to_json_format(value: str | None) -> str | None:
    """Attempts to parse a string as JSON and returns the parsed object or original string.

    Args:
        value: The string value to attempt JSON parsing on.

    Returns:
        The parsed JSON object if successful, otherwise the original string value.
    """
    try:
        if not value:
            return None
        return json.loads(value)
    except json.decoder.JSONDecodeError:
        return value


class HitlReader:
    """Handles reading and retrieving Human-In-The-Loop (HITL) data from UiPath services."""

    @classmethod
    async def read(cls, resume_trigger: UiPathResumeTrigger) -> Optional[str]:
        """Reads data from a UiPath resume trigger based on its type.

        This method handles different types of resume triggers (ACTION, JOB, API) and retrieves
        the appropriate data from UiPath services. For actions, it retrieves action data with
        optional escalation processing. For jobs, it retrieves job output and validates success.
        For API triggers, it retrieves the API payload.

        Args:
            resume_trigger: The UiPath resume trigger containing the trigger type and metadata.

        Returns:
            The retrieved data as a string, or None if no data is available.

        Raises:
            UiPathRuntimeError: If the job failed, API connection failed, trigger type is unknown,
                or HITL feedback retrieval failed.
        """
        uipath = UiPath()
        default_escalation = Escalation()
        match resume_trigger.trigger_type:
            case UiPathResumeTriggerType.ACTION:
                if resume_trigger.item_key:
                    action = await uipath.actions.retrieve_async(
                        resume_trigger.item_key,
                        app_folder_key=resume_trigger.folder_key,
                        app_folder_path=resume_trigger.folder_path,
                    )

                    if default_escalation.enabled:
                        return default_escalation.extract_response_value(action.data)

                    return action.data

            case UiPathResumeTriggerType.JOB:
                if resume_trigger.item_key:
                    job = await uipath.jobs.retrieve_async(
                        resume_trigger.item_key,
                        folder_key=resume_trigger.folder_key,
                        folder_path=resume_trigger.folder_path,
                    )
                    if (
                        job.state
                        and not job.state.lower()
                        == UiPathRuntimeStatus.SUCCESSFUL.value.lower()
                    ):
                        raise UiPathRuntimeError(
                            "INVOKED_PROCESS_FAILURE",
                            "Invoked process did not finish successfully.",
                            _try_convert_to_json_format(str(job.job_error or job.info))
                            or "Job error unavailable.",
                        )
                    output_data = await uipath.jobs.extract_output_async(job)
                    return _try_convert_to_json_format(output_data)

            case UiPathResumeTriggerType.API:
                if resume_trigger.api_resume and resume_trigger.api_resume.inbox_id:
                    try:
                        return await uipath.jobs.retrieve_api_payload_async(
                            resume_trigger.api_resume.inbox_id
                        )
                    except Exception as e:
                        raise UiPathRuntimeError(
                            "API_CONNECTION_ERROR",
                            "Failed to get trigger payload",
                            f"Error fetching API trigger payload for inbox {resume_trigger.api_resume.inbox_id}: {str(e)}",
                            UiPathErrorCategory.SYSTEM,
                        ) from e
            case _:
                raise UiPathRuntimeError(
                    "UNKNOWN_TRIGGER_TYPE",
                    "Unexpected trigger type received",
                    f"Trigger type :{type(resume_trigger.trigger_type)} is invalid",
                    UiPathErrorCategory.USER,
                )

        raise UiPathRuntimeError(
            "HITL_FEEDBACK_FAILURE",
            "Failed to receive payload from HITL action",
            detail="Failed to receive payload from HITL action",
            category=UiPathErrorCategory.SYSTEM,
        )


@dataclass
class HitlProcessor:
    """Processes events in a Human-(Robot/Agent)-In-The-Loop scenario.

    This class handles the creation and processing of HITL resume triggers for different
    types of UiPath operations including actions, jobs, and API calls. It determines the
    appropriate trigger type based on the input value and creates the corresponding
    resume trigger with proper configuration.

    Attributes:
        value: The input value to be processed, can be various UiPath model types or strings.
    """

    value: Any

    @cached_property
    def type(self) -> UiPathResumeTriggerType:
        """Determines the resume trigger type based on the input value.

        Analyzes the input value type and returns the corresponding UiPath resume trigger type.
        Actions (CreateAction, WaitAction) map to ACTION type, jobs (InvokeProcess, WaitJob)
        map to JOB type, and all other values default to API type.

        Returns:
            The appropriate UiPathResumeTriggerType based on the input value type.
        """
        if isinstance(self.value, CreateAction) or isinstance(self.value, WaitAction):
            return UiPathResumeTriggerType.ACTION
        if isinstance(self.value, InvokeProcess) or isinstance(self.value, WaitJob):
            return UiPathResumeTriggerType.JOB
        # default to API trigger
        return UiPathResumeTriggerType.API

    async def create_resume_trigger(self) -> UiPathResumeTrigger:
        """Creates a UiPath resume trigger based on the input value and its type.

        This method processes the input value and creates an appropriate resume trigger
        for HITL scenarios. It handles different input types:
        - Actions: Creates or references UiPath actions with folder information
        - Jobs: Invokes processes or references existing jobs with folder information
        - API: Creates API triggers with generated inbox IDs
        - String with escalation: Creates escalated actions

        Returns:
            A configured UiPathResumeTrigger ready for HITL processing.

        Raises:
            UiPathRuntimeError: If action/job creation fails, escalation fails, or an
                unknown model type is encountered.
            Exception: If any underlying UiPath service calls fail.
        """
        uipath = UiPath()
        default_escalation = Escalation()

        try:
            hitl_input = self.value
            resume_trigger = UiPathResumeTrigger(
                trigger_type=self.type, payload=serialize_object(hitl_input)
            )

            # check for default escalation config
            if default_escalation.enabled and isinstance(hitl_input, str):
                resume_trigger.trigger_type = UiPathResumeTriggerType.ACTION
                action = await default_escalation.create(hitl_input)
                if not action:
                    raise Exception("Failed to create default escalation")
                resume_trigger.item_key = action.key
                return resume_trigger

            match self.type:
                case UiPathResumeTriggerType.ACTION:
                    resume_trigger.folder_path = hitl_input.app_folder_path
                    resume_trigger.folder_key = hitl_input.app_folder_key
                    if isinstance(hitl_input, WaitAction):
                        resume_trigger.item_key = hitl_input.action.key
                    elif isinstance(hitl_input, CreateAction):
                        action = await uipath.actions.create_async(
                            title=hitl_input.title,
                            app_name=hitl_input.app_name if hitl_input.app_name else "",
                            app_folder_path=hitl_input.app_folder_path
                            if hitl_input.app_folder_path
                            else "",
                            app_folder_key=hitl_input.app_folder_key
                            if hitl_input.app_folder_key
                            else "",
                            app_key=hitl_input.app_key if hitl_input.app_key else "",
                            app_version=hitl_input.app_version
                            if hitl_input.app_version
                            else 1,
                            assignee=hitl_input.assignee if hitl_input.assignee else "",
                            data=hitl_input.data,
                        )
                        if not action:
                            raise Exception("Failed to create action")
                        resume_trigger.item_key = action.key

                case UiPathResumeTriggerType.JOB:
                    resume_trigger.folder_path = hitl_input.process_folder_path
                    resume_trigger.folder_key = hitl_input.process_folder_key
                    if isinstance(hitl_input, WaitJob):
                        resume_trigger.item_key = hitl_input.job.key
                    elif isinstance(hitl_input, InvokeProcess):
                        job = await uipath.processes.invoke_async(
                            name=hitl_input.name,
                            input_arguments=hitl_input.input_arguments,
                            folder_path=hitl_input.process_folder_path,
                            folder_key=hitl_input.process_folder_key,
                        )
                        if not job:
                            raise Exception("Failed to invoke process")
                        resume_trigger.item_key = job.key

                case UiPathResumeTriggerType.API:
                    resume_trigger.api_resume = UiPathApiTrigger(
                        inbox_id=str(uuid.uuid4()), request=serialize_object(hitl_input)
                    )
                case _:
                    raise UiPathRuntimeError(
                        "UNKNOWN_HITL_MODEL",
                        "Unexpected model received",
                        f"{type(hitl_input)} is not a valid Human(Robot/Agent)-In-The-Loop model",
                        UiPathErrorCategory.USER,
                    )
        except Exception as e:
            raise UiPathRuntimeError(
                "HITL_ACTION_CREATION_FAILED",
                "Failed to create HITL action",
                f"{str(e)}",
                UiPathErrorCategory.SYSTEM,
            ) from e

        return resume_trigger



"""The top-level event type representing an event in a conversation.

This is the root container for all other event subtypes (conversation start,
exchanges, messages, content, citations, tool calls, and async streams).
"""

from typing import Optional

from pydantic import BaseModel, ConfigDict, Field

from .async_stream import UiPathConversationAsyncInputStreamEvent
from .conversation import (
    UiPathConversationEndEvent,
    UiPathConversationStartedEvent,
    UiPathConversationStartEvent,
)
from .exchange import UiPathConversationExchangeEvent
from .meta import UiPathConversationMetaEvent
from .tool import UiPathConversationToolCallEvent


class UiPathConversationEvent(BaseModel):
    """The top-level event type representing an event in a conversation.

    This is the root container for all other event subtypes (conversation start,
    exchanges, messages, content, citations, tool calls, and async streams).
    """

    """A globally unique identifier for conversation to which the other sub-event and data properties apply."""
    conversation_id: str = Field(..., alias="conversationId")
    """Signals the start of an event stream concerning a conversation. This event does NOT necessarily mean this is a
    brand new conversation. It may be a continuation of an existing conversation.
    """
    start: Optional[UiPathConversationStartEvent] = None
    """Signals the acceptance of the start of a conversation."""
    started: Optional[UiPathConversationStartedEvent] = None
    """Signals the end of a conversation event stream. This does NOT mean the conversation is over. A new event stream for
    the conversation could be started in the future.
    """
    end: Optional[UiPathConversationEndEvent] = None
    """Encapsulates sub-events related to an exchange within a conversation."""
    exchange: Optional[UiPathConversationExchangeEvent] = None
    """Encapsulates sub-events related to an asynchronous input stream."""
    async_input_stream: Optional[UiPathConversationAsyncInputStreamEvent] = Field(
        None, alias="asyncInputStream"
    )
    """Optional async tool call sub-event. This feature is not supported by all LLMs. Most tool calls are scoped to a
    message, and use the toolCall and toolResult properties defined by the ConversationMessage type.
    """
    async_tool_call: Optional[UiPathConversationToolCallEvent] = Field(
        None, alias="asyncToolCall"
    )
    """Allows additional events to be sent in the context of the enclosing event stream."""
    meta_event: Optional[UiPathConversationMetaEvent] = Field(None, alias="metaEvent")

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)



import importlib.metadata
import inspect
import logging
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional

from ._utils._console import ConsoleLogger

logger = logging.getLogger(__name__)
console = ConsoleLogger()


@dataclass
class MiddlewareResult:
    should_continue: bool
    info_message: Optional[str] = None
    error_message: Optional[str] = None
    should_include_stacktrace: bool = False


MiddlewareFunc = Callable[..., MiddlewareResult]


class Middlewares:
    _middlewares: Dict[str, List[MiddlewareFunc]] = {
        "new": [],
        "init": [],
        "pack": [],
        "publish": [],
        "run": [],
        "dev": [],
        "invoke": [],
        "eval": [],
    }
    _plugins_loaded = False

    @classmethod
    def register(cls, command: str, middleware: MiddlewareFunc) -> None:
        """Register a middleware for a specific command."""
        if command not in cls._middlewares:
            cls._middlewares[command] = []
        cls._middlewares[command].append(middleware)
        logger.debug(
            f"Registered middleware for command '{command}': {middleware.__name__}"
        )

    @classmethod
    def get(cls, command: str) -> List[MiddlewareFunc]:
        """Get all middlewares for a specific command."""
        return cls._middlewares.get(command, [])

    @classmethod
    def next(cls, command: str, *args: Any, **kwargs: Any) -> MiddlewareResult:
        """Invoke middleware."""
        if not cls._plugins_loaded:
            cls.load_plugins()

        middlewares = cls.get(command)
        for middleware in middlewares:
            sig = inspect.signature(middleware)

            # handle older versions of plugins that don't support the new signature
            try:
                bound = sig.bind(*args, **kwargs)
                new_args = bound.args
                new_kwargs = bound.kwargs
            except TypeError:
                console.warning("Install the latest version for uipath packages")
                accepted_args = [
                    name
                    for name, param in sig.parameters.items()
                    if param.kind
                    in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD)
                ]

                trimmed_args = args[: len(accepted_args)]
                trimmed_kwargs = {k: v for k, v in kwargs.items() if k in accepted_args}

                new_args = trimmed_args
                new_kwargs = trimmed_kwargs

            try:
                result = middleware(*new_args, **new_kwargs)
                if not result.should_continue:
                    logger.debug(
                        f"Command '{command}' stopped by {middleware.__name__}"
                    )
                    return result
            except Exception as e:
                logger.error(f"Middleware {middleware.__name__} failed: {str(e)}")
                raise
        return MiddlewareResult(should_continue=True)

    @classmethod
    def clear(cls, command: Optional[str] = None) -> None:
        """Clear middlewares for a specific command or all middlewares if command is None."""
        if command:
            if command in cls._middlewares:
                cls._middlewares[command] = []
        else:
            for cmd in cls._middlewares:
                cls._middlewares[cmd] = []

    @classmethod
    def load_plugins(cls) -> None:
        """Load all middlewares registered via entry points."""
        if cls._plugins_loaded:
            return

        try:
            try:
                entry_points = importlib.metadata.entry_points()
                if hasattr(entry_points, "select"):
                    middlewares = list(entry_points.select(group="uipath.middlewares"))
                else:
                    middlewares = list(entry_points.get("uipath.middlewares", []))
            except Exception:
                middlewares = list(importlib.metadata.entry_points())  # type: ignore
                middlewares = [
                    ep for ep in middlewares if ep.group == "uipath.middlewares"
                ]

            if middlewares:
                logger.info(f"Found {len(middlewares)} middleware plugins")

                for entry_point in middlewares:
                    try:
                        register_func = entry_point.load()
                        register_func()
                        logger.info(f"Loaded middleware plugin: {entry_point.name}")
                    except Exception as e:
                        console.error(
                            f"Failed to load middleware plugin {entry_point.name}: {str(e)}",
                            include_traceback=True,
                        )
            else:
                logger.info("No middleware plugins found")

        except Exception as e:
            logger.error(f"No middleware plugins loaded: {str(e)}")
        finally:
            cls._plugins_loaded = True



# type: ignore
import json
import os
import uuid
import zipfile
from string import Template

import click

from uipath._cli._utils._constants import UIPATH_PROJECT_ID

from ..telemetry import track
from ..telemetry._constants import _PROJECT_KEY, _TELEMETRY_CONFIG_FILE
from ._utils._console import ConsoleLogger
from ._utils._project_files import (
    ensure_config_file,
    files_to_include,
    get_project_config,
    read_toml_project,
    validate_config,
)
from ._utils._uv_helpers import handle_uv_operations

console = ConsoleLogger()

schema = "https://cloud.uipath.com/draft/2024-12/entry-point"


def get_project_id() -> str:
    """Get project ID from telemetry file if it exists, otherwise generate a new one.

    Returns:
        Project ID string (either from telemetry file or newly generated).
    """
    # first check if this is a studio project
    if os.getenv(UIPATH_PROJECT_ID, None):
        return os.getenv(UIPATH_PROJECT_ID)

    telemetry_file = os.path.join(".uipath", _TELEMETRY_CONFIG_FILE)

    if os.path.exists(telemetry_file):
        try:
            with open(telemetry_file, "r") as f:
                telemetry_data = json.load(f)
                project_id = telemetry_data.get(_PROJECT_KEY)
                if project_id:
                    return project_id
        except (json.JSONDecodeError, IOError):
            pass

    return str(uuid.uuid4())


def get_project_version(directory):
    toml_path = os.path.join(directory, "pyproject.toml")
    if not os.path.exists(toml_path):
        console.warning("pyproject.toml not found. Using default version 0.0.1")
        return "0.0.1"
    toml_data = read_toml_project(toml_path)
    return toml_data["version"]


def validate_config_structure(config_data):
    required_fields = ["entryPoints"]
    for field in required_fields:
        if field not in config_data:
            console.error(f"uipath.json is missing the required field: {field}.")


def generate_operate_file(entryPoints, dependencies=None):
    project_id = get_project_id()

    first_entry = entryPoints[0]
    file_path = first_entry["filePath"]
    type = first_entry["type"]

    operate_json_data = {
        "$schema": schema,
        "projectId": project_id,
        "main": file_path,
        "contentType": type,
        "targetFramework": "Portable",
        "targetRuntime": "python",
        "runtimeOptions": {"requiresUserInteraction": False, "isAttended": False},
    }

    # Add dependencies if provided
    if dependencies:
        operate_json_data["dependencies"] = dependencies

    return operate_json_data


def generate_entrypoints_file(entryPoints):
    entrypoint_json_data = {
        "$schema": schema,
        "$id": "entry-points.json",
        "entryPoints": entryPoints,
    }

    return entrypoint_json_data


def generate_bindings_content():
    bindings_content = {"version": "2.0", "resources": []}

    return bindings_content


def generate_content_types_content():
    templates_path = os.path.join(
        os.path.dirname(__file__), "_templates", "[Content_Types].xml.template"
    )
    with open(templates_path, "r") as file:
        content_types_content = file.read()
    return content_types_content


def generate_nuspec_content(projectName, packageVersion, description, authors):
    variables = {
        "packageName": projectName,
        "packageVersion": packageVersion,
        "description": description,
        "authors": authors,
    }
    templates_path = os.path.join(
        os.path.dirname(__file__), "_templates", "package.nuspec.template"
    )
    with open(templates_path, "r", encoding="utf-8-sig") as f:
        content = f.read()
    return Template(content).substitute(variables)


def generate_rels_content(nuspecPath, psmdcpPath):
    # /package/services/metadata/core-properties/254324ccede240e093a925f0231429a0.psmdcp
    templates_path = os.path.join(
        os.path.dirname(__file__), "_templates", ".rels.template"
    )
    nuspecId = "R" + str(uuid.uuid4()).replace("-", "")[:16]
    psmdcpId = "R" + str(uuid.uuid4()).replace("-", "")[:16]
    variables = {
        "nuspecPath": nuspecPath,
        "nuspecId": nuspecId,
        "psmdcpPath": psmdcpPath,
        "psmdcpId": psmdcpId,
    }
    with open(templates_path, "r", encoding="utf-8-sig") as f:
        content = f.read()
    return Template(content).substitute(variables)


def generate_psmdcp_content(projectName, version, description, authors):
    templates_path = os.path.join(
        os.path.dirname(__file__), "_templates", ".psmdcp.template"
    )

    token = str(uuid.uuid4()).replace("-", "")[:32]
    random_file_name = f"{uuid.uuid4().hex[:16]}.psmdcp"
    variables = {
        "creator": authors,
        "description": description,
        "packageVersion": version,
        "projectName": projectName,
        "publicKeyToken": token,
    }
    with open(templates_path, "r", encoding="utf-8-sig") as f:
        content = f.read()

    return [random_file_name, Template(content).substitute(variables)]


def generate_package_descriptor_content(entryPoints):
    files = {
        "operate.json": "content/operate.json",
        "entry-points.json": "content/entry-points.json",
        "bindings.json": "content/bindings_v2.json",
    }

    for entry in entryPoints:
        files[entry["filePath"]] = entry["filePath"]

    package_descriptor_content = {
        "$schema": "https://cloud.uipath.com/draft/2024-12/package-descriptor",
        "files": files,
    }

    return package_descriptor_content


def is_venv_dir(d):
    return (
        os.path.exists(os.path.join(d, "Scripts", "activate"))
        if os.name == "nt"
        else os.path.exists(os.path.join(d, "bin", "activate"))
    )


def pack_fn(
    projectName,
    description,
    entryPoints,
    version,
    authors,
    directory,
    dependencies=None,
    include_uv_lock=True,
):
    operate_file = generate_operate_file(entryPoints, dependencies)
    entrypoints_file = generate_entrypoints_file(entryPoints)

    # Get bindings from uipath.json if available
    config_path = os.path.join(directory, "uipath.json")
    if not os.path.exists(config_path):
        console.error("uipath.json not found, please run `uipath init`.")

    with open(config_path, "r") as f:
        config_data = json.load(f)
        if "bindings" in config_data:
            bindings_content = config_data["bindings"]
        else:
            bindings_content = generate_bindings_content()

    content_types_content = generate_content_types_content()
    [psmdcp_file_name, psmdcp_content] = generate_psmdcp_content(
        projectName, version, description, authors
    )
    nuspec_content = generate_nuspec_content(projectName, version, description, authors)
    rels_content = generate_rels_content(
        f"/{projectName}.nuspec",
        f"/package/services/metadata/core-properties/{psmdcp_file_name}",
    )
    package_descriptor_content = generate_package_descriptor_content(entryPoints)

    # Create .uipath directory if it doesn't exist
    os.makedirs(".uipath", exist_ok=True)

    with zipfile.ZipFile(
        f".uipath/{projectName}.{version}.nupkg", "w", zipfile.ZIP_DEFLATED
    ) as z:
        # Add metadata files
        z.writestr(
            f"./package/services/metadata/core-properties/{psmdcp_file_name}",
            psmdcp_content,
        )
        z.writestr("[Content_Types].xml", content_types_content)
        z.writestr(
            "content/package-descriptor.json",
            json.dumps(package_descriptor_content, indent=4),
        )
        z.writestr("content/operate.json", json.dumps(operate_file, indent=4))
        z.writestr("content/entry-points.json", json.dumps(entrypoints_file, indent=4))
        z.writestr("content/bindings_v2.json", json.dumps(bindings_content, indent=4))
        z.writestr(f"{projectName}.nuspec", nuspec_content)
        z.writestr("_rels/.rels", rels_content)

        files = files_to_include(config_data, directory, include_uv_lock)

        for file in files:
            if file.is_binary:
                # Read binary files in binary mode
                with open(file.file_path, "rb") as f:
                    z.writestr(f"content/{file.relative_path}", f.read())
            else:
                try:
                    # Try UTF-8 first
                    with open(file.file_path, "r", encoding="utf-8") as f:
                        z.writestr(f"content/{file.relative_path}", f.read())
                except UnicodeDecodeError:
                    # If UTF-8 fails, try with utf-8-sig (for files with BOM)
                    try:
                        with open(file.file_path, "r", encoding="utf-8-sig") as f:
                            z.writestr(f"content/{file.relative_path}", f.read())
                    except UnicodeDecodeError:
                        # If that also fails, try with latin-1 as a fallback
                        with open(file.file_path, "r", encoding="latin-1") as f:
                            z.writestr(f"content/{file.relative_path}", f.read())


def display_project_info(config):
    max_label_length = max(
        len(label) for label in ["Name", "Version", "Description", "Authors"]
    )

    max_length = 100
    description = config["description"]
    if len(description) >= max_length:
        description = description[: max_length - 3] + " ..."

    console.log(f"{'Name'.ljust(max_label_length)}: {config['project_name']}")
    console.log(f"{'Version'.ljust(max_label_length)}: {config['version']}")
    console.log(f"{'Description'.ljust(max_label_length)}: {description}")
    console.log(f"{'Authors'.ljust(max_label_length)}: {config['authors']}")


@click.command()
@click.argument(
    "root", type=click.Path(exists=True, file_okay=False, dir_okay=True), default="."
)
@click.option(
    "--nolock",
    is_flag=True,
    help="Skip running uv lock and exclude uv.lock from the package",
)
@track
def pack(root, nolock):
    """Pack the project."""
    version = get_project_version(root)

    ensure_config_file(root)
    config = get_project_config(root)
    validate_config(config)

    with console.spinner("Packaging project ..."):
        try:
            # Handle uv operations before packaging, unless nolock is specified
            if not nolock:
                handle_uv_operations(root)

            pack_fn(
                config["project_name"],
                config["description"],
                config["entryPoints"],
                version or config["version"],
                config["authors"],
                root,
                config.get("dependencies"),
                include_uv_lock=not nolock,
            )
            display_project_info(config)
            console.success("Project successfully packaged.")

        except Exception as e:
            console.error(
                f"Failed to create package {config['project_name']}.{version or config['version']}: {str(e)}"
            )


if __name__ == "__main__":
    pack()



import asyncio
import json
import traceback
from datetime import datetime
from os import environ as env
from pathlib import Path
from typing import Any, Dict, cast
from uuid import uuid4

import pyperclip  # type: ignore[import-untyped]
from rich.traceback import Traceback
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal
from textual.widgets import Button, Footer, Input, ListView, RichLog

from uipath.agent.conversation import (
    UiPathConversationContentPart,
    UiPathConversationEvent,
    UiPathConversationMessage,
    UiPathInlineValue,
)

from ..._runtime._contracts import (
    UiPathErrorContract,
    UiPathRuntimeContext,
    UiPathRuntimeError,
    UiPathRuntimeFactory,
    UiPathRuntimeStatus,
)
from ._components._details import RunDetailsPanel
from ._components._history import RunHistoryPanel
from ._components._new import NewRunPanel
from ._models._execution import ExecutionRun
from ._models._messages import LogMessage, TraceMessage
from ._utils._chat import RunContextChatHandler, build_user_message_event
from ._utils._exporter import RunContextExporter
from ._utils._logger import RunContextLogHandler


class UiPathDevTerminal(App[Any]):
    """UiPath debugging terminal interface."""

    TITLE = "UiPath Debugging Terminal"
    SUB_TITLE = "Interactive debugging interface for UiPath Python projects"
    CSS_PATH = Path(__file__).parent / "_styles" / "terminal.tcss"

    BINDINGS = [
        Binding("q", "quit", "Quit"),
        Binding("n", "new_run", "New"),
        Binding("r", "execute_run", "Run"),
        Binding("c", "copy", "Copy"),
        Binding("h", "clear_history", "Clear History"),
        Binding("escape", "cancel", "Cancel"),
    ]

    def __init__(
        self,
        runtime_factory: UiPathRuntimeFactory[Any, Any],
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.initial_entrypoint: str = "main.py"
        self.initial_input: str = '{\n  "message": "Hello World"\n}'
        self.runs: Dict[str, ExecutionRun] = {}
        self.runtime_factory = runtime_factory
        self.runtime_factory.add_span_exporter(
            RunContextExporter(
                on_trace=self._handle_trace_message,
                on_log=self._handle_log_message,
            ),
            batch=False,
        )

    def compose(self) -> ComposeResult:
        with Horizontal():
            # Left sidebar - run history
            with Container(classes="run-history"):
                yield RunHistoryPanel(id="history-panel")

            # Main content area
            with Container(classes="main-content"):
                # New run panel (initially visible)
                yield NewRunPanel(
                    id="new-run-panel",
                    classes="new-run-panel",
                )

                # Run details panel (initially hidden)
                yield RunDetailsPanel(id="details-panel", classes="hidden")

        yield Footer()

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button press events."""
        if event.button.id == "new-run-btn":
            await self.action_new_run()
        elif event.button.id == "execute-btn":
            await self.action_execute_run()
        elif event.button.id == "cancel-btn":
            await self.action_cancel()

    async def on_list_view_selected(self, event: ListView.Selected) -> None:
        """Handle run selection from history."""
        if event.list_view.id == "run-list" and event.item:
            run_id = getattr(event.item, "run_id", None)
            if run_id:
                history_panel = self.query_one("#history-panel", RunHistoryPanel)
                run = history_panel.get_run_by_id(run_id)
                if run:
                    self._show_run_details(run)

    @on(Input.Submitted, "#chat-input")
    async def handle_chat_input(self, event: Input.Submitted) -> None:
        """Handle user submitting text into the chat."""
        user_text = event.value.strip()
        if not user_text:
            return

        details_panel = self.query_one("#details-panel", RunDetailsPanel)
        if details_panel and details_panel.current_run:
            status = details_panel.current_run.status
            if status == "running":
                self.app.notify(
                    "Wait for agent response...", timeout=1.5, severity="warning"
                )
                return
            self._handle_chat_event(
                build_user_message_event(
                    user_text=user_text,
                    conversation_id=details_panel.current_run.id,
                ),
                details_panel.current_run.id,
            )
            if details_panel.current_run.status == "suspended":
                details_panel.current_run.resume_data = user_text
            else:
                details_panel.current_run.input_data = UiPathConversationMessage(
                    message_id=str(uuid4()),
                    created_at=datetime.now().isoformat(),
                    updated_at=datetime.now().isoformat(),
                    content_parts=[
                        UiPathConversationContentPart(
                            content_part_id=str(uuid4()),
                            mime_type="text/plain",
                            data=UiPathInlineValue(inline=user_text),
                        )
                    ],
                    role="user",
                )
            asyncio.create_task(self._execute_runtime(details_panel.current_run))
            event.input.clear()

    async def action_new_run(self) -> None:
        """Show new run panel."""
        new_panel = self.query_one("#new-run-panel")
        details_panel = self.query_one("#details-panel")

        new_panel.remove_class("hidden")
        details_panel.add_class("hidden")

    async def action_cancel(self) -> None:
        """Cancel and return to new run view."""
        await self.action_new_run()

    async def action_execute_run(self) -> None:
        """Execute a new run with UiPath runtime."""
        new_run_panel = self.query_one("#new-run-panel", NewRunPanel)
        entrypoint, input_data, conversational = new_run_panel.get_input_values()

        if not entrypoint:
            return

        input: Dict[str, Any] = {}
        try:
            input = json.loads(input_data)
        except json.JSONDecodeError:
            return

        run = ExecutionRun(entrypoint, input, conversational)

        self.runs[run.id] = run

        self._add_run_in_history(run)

        self._show_run_details(run)

        if not run.conversational:
            asyncio.create_task(self._execute_runtime(run))
        else:
            self._focus_chat_input()

    async def action_clear_history(self) -> None:
        """Clear run history."""
        history_panel = self.query_one("#history-panel", RunHistoryPanel)
        history_panel.clear_runs()
        await self.action_new_run()

    def action_copy(self) -> None:
        """Copy content of currently focused RichLog to clipboard and notify."""
        focused = self.app.focused
        if isinstance(focused, RichLog):
            clipboard_text = "\n".join(line.text for line in focused.lines)
            pyperclip.copy(clipboard_text)
            self.app.notify("Copied to clipboard!", timeout=1.5)
        else:
            self.app.notify("Nothing to copy here.", timeout=1.5, severity="warning")

    async def _execute_runtime(self, run: ExecutionRun):
        """Execute the script using UiPath runtime."""
        try:
            context: UiPathRuntimeContext = self.runtime_factory.new_context(
                entrypoint=run.entrypoint,
                trace_id=str(uuid4()),
                execution_id=run.id,
                is_conversational=run.conversational,
                logs_min_level=env.get("LOG_LEVEL", "INFO"),
                log_handler=RunContextLogHandler(
                    run_id=run.id, callback=self._handle_log_message
                ),
                chat_handler=RunContextChatHandler(
                    run_id=run.id, callback=self._handle_chat_event
                ),
            )

            if run.status == "suspended":
                context.input_json = run.resume_data
                context.resume = True
                self._add_info_log(run, f"Resuming execution: {run.entrypoint}")
            else:
                if run.conversational:
                    context.input_message = cast(
                        UiPathConversationMessage, run.input_data
                    )
                else:
                    context.input_json = run.input_data
                self._add_info_log(run, f"Starting execution: {run.entrypoint}")

            run.status = "running"
            run.start_time = datetime.now()

            result = await self.runtime_factory.execute_in_root_span(context)

            if result is not None:
                if (
                    result.status == UiPathRuntimeStatus.SUSPENDED.value
                    and result.resume
                ):
                    run.status = "suspended"
                else:
                    run.output_data = result.output
                    run.status = "completed"
                if run.output_data:
                    self._add_info_log(run, f"Execution result: {run.output_data}")

            self._add_info_log(run, "✅ Execution completed successfully")
            run.end_time = datetime.now()

        except UiPathRuntimeError as e:
            self._add_error_log(run)
            run.status = "failed"
            run.end_time = datetime.now()
            run.error = e.error_info

        except Exception as e:
            self._add_error_log(run)
            run.status = "failed"
            run.end_time = datetime.now()
            run.error = UiPathErrorContract(
                code="Unknown", title=str(e), detail=traceback.format_exc()
            )

        self._update_run_in_history(run)
        self._update_run_details(run)

    def _show_run_details(self, run: ExecutionRun):
        """Show details panel for a specific run."""
        # Hide new run panel, show details panel
        new_panel = self.query_one("#new-run-panel")
        details_panel = self.query_one("#details-panel", RunDetailsPanel)

        new_panel.add_class("hidden")
        details_panel.remove_class("hidden")

        # Populate the details panel with run data
        details_panel.update_run(run)

    def _focus_chat_input(self):
        """Focus the chat input box."""
        details_panel = self.query_one("#details-panel", RunDetailsPanel)
        details_panel.switch_tab("chat-tab")
        chat_input = details_panel.query_one("#chat-input", Input)
        chat_input.focus()

    def _add_run_in_history(self, run: ExecutionRun):
        """Add run to history panel."""
        history_panel = self.query_one("#history-panel", RunHistoryPanel)
        history_panel.add_run(run)

    def _update_run_in_history(self, run: ExecutionRun):
        """Update run display in history panel."""
        history_panel = self.query_one("#history-panel", RunHistoryPanel)
        history_panel.update_run(run)

    def _update_run_details(self, run: ExecutionRun):
        """Update the displayed run information."""
        details_panel = self.query_one("#details-panel", RunDetailsPanel)
        details_panel.update_run_details(run)

    def _handle_trace_message(self, trace_msg: TraceMessage):
        """Handle trace message from exporter."""
        run = self.runs[trace_msg.run_id]
        for i, existing_trace in enumerate(run.traces):
            if existing_trace.span_id == trace_msg.span_id:
                run.traces[i] = trace_msg
                break
        else:
            run.traces.append(trace_msg)

        details_panel = self.query_one("#details-panel", RunDetailsPanel)
        details_panel.add_trace(trace_msg)

    def _handle_log_message(self, log_msg: LogMessage):
        """Handle log message from exporter."""
        self.runs[log_msg.run_id].logs.append(log_msg)
        details_panel = self.query_one("#details-panel", RunDetailsPanel)
        details_panel.add_log(log_msg)

    def _handle_chat_event(
        self, event: UiPathConversationEvent, execution_id: str
    ) -> None:
        updated_chat_message = self.runs[execution_id].add_message(event)
        if not updated_chat_message:
            return
        details_panel = self.app.query_one("#details-panel", RunDetailsPanel)
        details_panel.add_chat_message(event, updated_chat_message, execution_id)

    def _add_info_log(self, run: ExecutionRun, message: str):
        """Add info log to run."""
        timestamp = datetime.now()
        log_msg = LogMessage(run.id, "INFO", message, timestamp)
        self._handle_log_message(log_msg)

    def _add_error_log(self, run: ExecutionRun):
        """Add error log to run."""
        timestamp = datetime.now()
        tb = Traceback(
            show_locals=False,
            max_frames=4,
        )
        log_msg = LogMessage(run.id, "ERROR", tb, timestamp)
        self._handle_log_message(log_msg)



import asyncio
import json
import os
import webbrowser
from socket import AF_INET, SOCK_STREAM, error, socket
from typing import Optional
from urllib.parse import urlparse

from uipath._cli._auth._auth_server import HTTPServer
from uipath._cli._auth._client_credentials import ClientCredentialsService
from uipath._cli._auth._oidc_utils import OidcUtils
from uipath._cli._auth._portal_service import PortalService, select_tenant
from uipath._cli._auth._url_utils import set_force_flag
from uipath._cli._auth._utils import update_auth_file, update_env_file
from uipath._cli._utils._console import ConsoleLogger


class AuthService:
    def __init__(
        self,
        environment: str,
        *,
        force: bool,
        client_id: Optional[str],
        client_secret: Optional[str],
        base_url: Optional[str],
        scope: Optional[str],
    ):
        self._force = force
        self._console = ConsoleLogger()
        self._domain = self._get_domain(environment)
        self._client_id = client_id
        self._client_secret = client_secret
        self._base_url = base_url
        self._scope = scope
        set_force_flag(self._force)

    def _get_domain(self, environment: str) -> str:
        # only search env var if not force authentication
        if not self._force:
            uipath_url = os.getenv("UIPATH_URL")
            if uipath_url and environment == "cloud":  # "cloud" is the default
                parsed_url = urlparse(uipath_url)
                if parsed_url.scheme and parsed_url.netloc:
                    environment = f"{parsed_url.scheme}://{parsed_url.netloc}"
                else:
                    self._console.error(
                        f"Malformed UIPATH_URL: '{uipath_url}'. Please ensure it includes both scheme and netloc (e.g., 'https://cloud.uipath.com')."
                    )
        return environment

    def authenticate(self) -> None:
        if self._client_id and self._client_secret:
            self._authenticate_client_credentials()
            return

        self._authenticate_authorization_code()

    def _authenticate_client_credentials(self) -> None:
        if not self._base_url:
            self._console.error(
                "--base-url is required when using client credentials authentication."
            )
            return
        self._console.hint("Using client credentials authentication.")
        credentials_service = ClientCredentialsService(self._base_url)
        credentials_service.authenticate(
            self._client_id,  # type: ignore
            self._client_secret,  # type: ignore
            self._scope,
        )

    def _authenticate_authorization_code(self) -> None:
        with PortalService(self._domain) as portal_service:
            if not self._force:
                # use existing env vars
                if (
                    os.getenv("UIPATH_URL")
                    and os.getenv("UIPATH_TENANT_ID")
                    and os.getenv("UIPATH_ORGANIZATION_ID")
                ):
                    try:
                        portal_service.ensure_valid_token()
                        return
                    except Exception:
                        self._console.error(
                            "Authentication token is invalid. Please reauthenticate using the '--force' flag.",
                        )
            auth_url, code_verifier, state = OidcUtils.get_auth_url(self._domain)
            webbrowser.open(auth_url, 1)
            auth_config = OidcUtils.get_auth_config()

            self._console.link(
                "If a browser window did not open, please open the following URL in your browser:",
                auth_url,
            )
            server = HTTPServer(port=auth_config["port"])
            token_data = asyncio.run(server.start(state, code_verifier, self._domain))

            if not token_data:
                self._console.error(
                    "Authentication failed. Please try again.",
                )

            portal_service.update_token_data(token_data)
            update_auth_file(token_data)
            access_token = token_data["access_token"]
            update_env_file({"UIPATH_ACCESS_TOKEN": access_token})

            tenants_and_organizations = portal_service.get_tenants_and_organizations()
            base_url = select_tenant(self._domain, tenants_and_organizations)
            try:
                portal_service.post_auth(base_url)
            except Exception:
                self._console.error(
                    "Could not prepare the environment. Please try again.",
                )

    def set_port(self):
        def is_port_in_use(target_port: int) -> bool:
            with socket(AF_INET, SOCK_STREAM) as s:
                try:
                    s.bind(("localhost", target_port))
                    s.close()
                    return False
                except error:
                    return True

        auth_config = OidcUtils.get_auth_config()
        port = int(auth_config.get("port", 8104))
        port_option_one = int(auth_config.get("portOptionOne", 8104))  # type: ignore
        port_option_two = int(auth_config.get("portOptionTwo", 8055))  # type: ignore
        port_option_three = int(auth_config.get("portOptionThree", 42042))  # type: ignore
        if is_port_in_use(port):
            if is_port_in_use(port_option_one):
                if is_port_in_use(port_option_two):
                    if is_port_in_use(port_option_three):
                        self._console.error(
                            "All configured ports are in use. Please close applications using ports or configure different ports."
                        )
                    else:
                        port = port_option_three
                else:
                    port = port_option_two
            else:
                port = port_option_one
        auth_config["port"] = port
        with open(
            os.path.join(os.path.dirname(__file__), "..", "auth_config.json"), "w"
        ) as f:
            json.dump(auth_config, f)



from langgraph.graph import START, StateGraph, END
from pydantic import BaseModel
import os

class GraphState(BaseModel):
    topic: str

class GraphOutput(BaseModel):
    report: str

async def generate_report(state: GraphState) -> GraphOutput:
    report = "Just a sanity report"
    return GraphOutput(report=report)

builder = StateGraph(GraphState, output=GraphOutput)

builder.add_node("generate_report", generate_report)

builder.add_edge(START, "generate_report")
builder.add_edge("generate_report", END)

graph = builder.compile()



from pydantic import BaseModel

from uipath.eval.models.models import EvaluatorCategory, EvaluatorType


class EvaluatorBaseParams(BaseModel):
    """Parameters for initializing the base evaluator."""

    id: str
    category: EvaluatorCategory
    evaluator_type: EvaluatorType
    name: str
    description: str
    created_at: str
    updated_at: str
    target_output_key: str



import base64
import hashlib
import json
import os
from urllib.parse import urlencode

from ._models import AuthConfig
from ._url_utils import build_service_url


def generate_code_verifier_and_challenge():
    """Generate PKCE code verifier and challenge."""
    code_verifier = base64.urlsafe_b64encode(os.urandom(32)).decode("utf-8").rstrip("=")

    code_challenge_bytes = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = (
        base64.urlsafe_b64encode(code_challenge_bytes).decode("utf-8").rstrip("=")
    )

    return code_verifier, code_challenge


def get_state_param() -> str:
    return base64.urlsafe_b64encode(os.urandom(32)).decode("utf-8").rstrip("=")


class OidcUtils:
    @classmethod
    def get_auth_config(cls) -> AuthConfig:
        with open(
            os.path.join(os.path.dirname(__file__), "auth_config.json"), "r"
        ) as f:
            auth_config = json.load(f)

        port = auth_config.get("port", 8104)

        redirect_uri = auth_config["redirect_uri"].replace(
            "__PY_REPLACE_PORT__", str(port)
        )

        return AuthConfig(
            client_id=auth_config["client_id"],
            redirect_uri=redirect_uri,
            scope=auth_config["scope"],
            port=port,
        )

    @classmethod
    def get_auth_url(cls, domain: str) -> tuple[str, str, str]:
        """Get the authorization URL for OAuth2 PKCE flow.

        Args:
            domain (str): The UiPath domain to authenticate against (e.g. 'alpha', 'cloud')

        Returns:
            tuple[str, str]: A tuple containing:
                - The authorization URL with query parameters
                - The code verifier for PKCE flow
        """
        code_verifier, code_challenge = generate_code_verifier_and_challenge()
        auth_config = cls.get_auth_config()
        state = get_state_param()
        query_params = {
            "client_id": auth_config["client_id"],
            "redirect_uri": auth_config["redirect_uri"],
            "response_type": "code",
            "scope": auth_config["scope"],
            "state": state,
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
        }

        query_string = urlencode(query_params)
        url = build_service_url(domain, f"/identity_/connect/authorize?{query_string}")
        return url, code_verifier, state



# type: ignore
import ast
import asyncio
import os
from typing import List, Optional

import click

from uipath._cli._evals._runtime import UiPathEvalContext, UiPathEvalRuntime
from uipath._cli._runtime._contracts import (
    UiPathRuntimeContext,
    UiPathRuntimeFactory,
)
from uipath._cli._runtime._runtime import UiPathScriptRuntime
from uipath._cli.middlewares import Middlewares
from uipath.eval._helpers import auto_discover_entrypoint

from .._utils.constants import ENV_JOB_ID
from ..telemetry import track
from ._utils._console import ConsoleLogger
from ._utils._eval_set import EvalHelpers

console = ConsoleLogger()


class LiteralOption(click.Option):
    def type_cast_value(self, ctx, value):
        try:
            return ast.literal_eval(value)
        except Exception as e:
            raise click.BadParameter(value) from e


@click.command()
@click.argument("entrypoint", required=False)
@click.argument("eval_set", required=False)
@click.option("--eval-ids", cls=LiteralOption, default="[]")
@click.option(
    "--no-report",
    is_flag=True,
    help="Do not report the evaluation results",
    default=False,
)
@click.option(
    "--workers",
    type=int,
    default=8,
    help="Number of parallel workers for running evaluations (default: 8)",
)
@click.option(
    "--output-file",
    required=False,
    type=click.Path(exists=False),
    help="File path where the output will be written",
)
@track(when=lambda *_a, **_kw: os.getenv(ENV_JOB_ID) is None)
def eval(
    entrypoint: Optional[str],
    eval_set: Optional[str],
    eval_ids: List[str],
    no_report: bool,
    workers: int,
    output_file: Optional[str],
) -> None:
    """Run an evaluation set against the agent.

    Args:
        entrypoint: Path to the agent script to evaluate (optional, will auto-discover if not specified)
        eval_set: Path to the evaluation set JSON file (optional, will auto-discover if not specified)
        eval_ids: Optional list of evaluation IDs
        workers: Number of parallel workers for running evaluations
        no_report: Do not report the evaluation results
    """
    result = Middlewares.next(
        "eval",
        entrypoint,
        eval_set,
        eval_ids,
        no_report=no_report,
        workers=workers,
        execution_output_file=output_file,
    )

    if result.error_message:
        console.error(result.error_message)

    if result.should_continue:

        def generate_runtime_context(**context_kwargs) -> UiPathRuntimeContext:
            runtime_context = UiPathRuntimeContext.with_defaults(**context_kwargs)
            runtime_context.entrypoint = runtime_entrypoint
            return runtime_context

        eval_context = UiPathEvalContext.with_defaults(
            execution_output_file=output_file
        )

        eval_context.no_report = no_report
        eval_context.workers = workers
        eval_context.eval_set = eval_set or EvalHelpers.auto_discover_eval_set()
        eval_context.eval_ids = eval_ids

        runtime_entrypoint = entrypoint or auto_discover_entrypoint()

        try:
            runtime_factory = UiPathRuntimeFactory(
                UiPathScriptRuntime,
                UiPathRuntimeContext,
                context_generator=generate_runtime_context,
            )

            async def execute():
                async with UiPathEvalRuntime.from_eval_context(
                    factory=runtime_factory, context=eval_context
                ) as eval_runtime:
                    await eval_runtime.execute()

            asyncio.run(execute())
        except Exception as e:
            console.error(
                f"Error: Unexpected error occurred - {str(e)}", include_traceback=True
            )

    console.success("Evaluation completed successfully")


if __name__ == "__main__":
    eval()



from .actions_service import ActionsService
from .api_client import ApiClient
from .assets_service import AssetsService
from .attachments_service import AttachmentsService
from .buckets_service import BucketsService
from .connections_service import ConnectionsService
from .context_grounding_service import ContextGroundingService
from .entities_service import EntitiesService
from .folder_service import FolderService
from .jobs_service import JobsService
from .llm_gateway_service import UiPathLlmChatService, UiPathOpenAIService
from .processes_service import ProcessesService
from .queues_service import QueuesService

__all__ = [
    "ActionsService",
    "AssetsService",
    "AttachmentsService",
    "BucketsService",
    "ConnectionsService",
    "ContextGroundingService",
    "ProcessesService",
    "ApiClient",
    "QueuesService",
    "JobsService",
    "UiPathOpenAIService",
    "UiPathLlmChatService",
    "FolderService",
    "EntitiesService",
]



import json
from collections import defaultdict
from pathlib import Path
from time import time
from typing import Any, Dict, Generic, List, Optional, Sequence, TypeVar

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult

from ...eval.evaluators import BaseEvaluator
from ...eval.models import EvaluationResult
from ...eval.models.models import AgentExecution
from .._runtime._contracts import (
    UiPathBaseRuntime,
    UiPathRuntimeContext,
    UiPathRuntimeFactory,
    UiPathRuntimeResult,
    UiPathRuntimeStatus,
)
from .._utils._eval_set import EvalHelpers
from ._evaluator_factory import EvaluatorFactory
from ._models._evaluation_set import EvaluationItem, EvaluationSet
from ._models._output import (
    EvaluationResultDto,
    EvaluationRunResult,
    EvaluationRunResultDto,
    UiPathEvalOutput,
    UiPathEvalRunExecutionOutput,
)

T = TypeVar("T", bound=UiPathBaseRuntime)
C = TypeVar("C", bound=UiPathRuntimeContext)


class ExecutionSpanExporter(SpanExporter):
    """Custom exporter that stores spans grouped by execution ids."""

    def __init__(self):
        # { execution_id -> list of spans }
        self._spans: Dict[str, List[ReadableSpan]] = defaultdict(list)

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for span in spans:
            if span.attributes is not None:
                exec_id = span.attributes.get("execution.id")
                if exec_id is not None and isinstance(exec_id, str):
                    self._spans[exec_id].append(span)

        return SpanExportResult.SUCCESS

    def get_spans(self, execution_id: str) -> List[ReadableSpan]:
        """Retrieve spans for a given execution id."""
        return self._spans.get(execution_id, [])

    def clear(self, execution_id: Optional[str] = None) -> None:
        """Clear stored spans for one or all executions."""
        if execution_id:
            self._spans.pop(execution_id, None)
        else:
            self._spans.clear()

    def shutdown(self) -> None:
        self.clear()


class UiPathEvalContext(UiPathRuntimeContext):
    """Context used for evaluation runs."""

    no_report: Optional[bool] = False
    workers: Optional[int] = 1
    eval_set: Optional[str] = None
    eval_ids: Optional[List[str]] = None


class UiPathEvalRuntime(UiPathBaseRuntime, Generic[T, C]):
    """Specialized runtime for evaluation runs, with access to the factory."""

    def __init__(self, context: UiPathEvalContext, factory: UiPathRuntimeFactory[T, C]):
        super().__init__(context)
        self.context: UiPathEvalContext = context
        self.factory: UiPathRuntimeFactory[T, C] = factory
        self.span_exporter: ExecutionSpanExporter = ExecutionSpanExporter()
        self.factory.add_span_exporter(self.span_exporter)

    @classmethod
    def from_eval_context(
        cls,
        context: UiPathEvalContext,
        factory: UiPathRuntimeFactory[T, C],
    ) -> "UiPathEvalRuntime[T, C]":
        return cls(context, factory)

    async def execute(self) -> Optional[UiPathRuntimeResult]:
        """Evaluation logic. Can spawn other runtimes through the factory."""
        if self.context.eval_set is None:
            raise ValueError("eval_set must be provided for evaluation runs")

        evaluation_set = EvalHelpers.load_eval_set(
            self.context.eval_set, self.context.eval_ids
        )
        evaluators = self._load_evaluators(evaluation_set)
        results = UiPathEvalOutput(
            evaluation_set_name=evaluation_set.name, score=0, evaluation_set_results=[]
        )
        for eval_item in evaluation_set.evaluations:
            evaluation_run_results = EvaluationRunResult(
                evaluation_name=eval_item.name, evaluation_run_results=[]
            )

            results.evaluation_set_results.append(evaluation_run_results)
            agent_execution_output = await self.execute_runtime(eval_item)
            # we run each evaluator on the agent_output
            for evaluator in evaluators:
                evaluation_result = await self.run_evaluator(
                    evaluator=evaluator,
                    execution_output=agent_execution_output,
                    eval_item=eval_item,
                )
                evaluation_run_results.evaluation_run_results.append(
                    EvaluationRunResultDto(
                        evaluator_name=evaluator.name,
                        result=EvaluationResultDto.from_evaluation_result(
                            evaluation_result
                        ),
                    )
                )

        results.compute_average_score()
        self.context.result = UiPathRuntimeResult(
            output={**results.model_dump(by_alias=True)},
            status=UiPathRuntimeStatus.SUCCESSFUL,
        )

        return self.context.result

    async def execute_runtime(
        self, eval_item: EvaluationItem
    ) -> UiPathEvalRunExecutionOutput:
        runtime_context: C = self.factory.new_context(
            execution_id=eval_item.id,
            input_json=eval_item.inputs,
            is_eval_run=True,
        )
        start_time = time()
        result = await self.factory.execute_in_root_span(
            runtime_context, root_span=eval_item.name
        )
        end_time = time()
        if runtime_context.execution_id is None:
            raise ValueError("execution_id must be set for eval runs")

        spans = self.span_exporter.get_spans(runtime_context.execution_id)
        self.span_exporter.clear(runtime_context.execution_id)

        if result is None:
            raise ValueError("Execution result cannot be None for eval runs")

        return UiPathEvalRunExecutionOutput(
            execution_time=end_time - start_time,
            spans=spans,
            result=result,
        )

    async def run_evaluator(
        self,
        evaluator: BaseEvaluator[Any],
        execution_output: UiPathEvalRunExecutionOutput,
        eval_item: EvaluationItem,
    ) -> EvaluationResult:
        agent_execution = AgentExecution(
            agent_input=eval_item.inputs,
            agent_output=execution_output.result.output or {},
            agent_trace=execution_output.spans,
        )

        result = await evaluator.evaluate(
            agent_execution=agent_execution,
            # at the moment evaluation_criteria is always the expected output
            evaluation_criteria=eval_item.expected_output,
        )

        return result

    def _load_evaluators(
        self, evaluation_set: EvaluationSet
    ) -> List[BaseEvaluator[Any]]:
        """Load evaluators referenced by the evaluation set."""
        evaluators = []
        evaluators_dir = Path(self.context.eval_set).parent.parent / "evaluators"  # type: ignore
        evaluator_refs = set(evaluation_set.evaluator_refs)
        found_evaluator_ids = set()

        for file in evaluators_dir.glob("*.json"):
            try:
                with open(file, "r", encoding="utf-8") as f:
                    data = json.load(f)
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"Invalid JSON in evaluator file '{file}': {str(e)}. "
                    f"Please check the file for syntax errors."
                ) from e

            try:
                evaluator_id = data.get("id")
                if evaluator_id in evaluator_refs:
                    evaluator = EvaluatorFactory.create_evaluator(data)
                    evaluators.append(evaluator)
                    found_evaluator_ids.add(evaluator_id)
            except Exception as e:
                raise ValueError(
                    f"Failed to create evaluator from file '{file}': {str(e)}. "
                    f"Please verify the evaluator configuration."
                ) from e

        missing_evaluators = evaluator_refs - found_evaluator_ids
        if missing_evaluators:
            raise ValueError(
                f"Could not find the following evaluators: {missing_evaluators}"
            )

        return evaluators

    async def cleanup(self) -> None:
        """Cleanup runtime resources."""
        pass

    async def validate(self) -> None:
        """Cleanup runtime resources."""
        pass



from typing import Optional

from typing_extensions import deprecated

from uipath.tracing._traced import traced

from .._config import Config
from .._execution_context import ExecutionContext
from .._utils import Endpoint, RequestSpec
from ._base_service import BaseService


class FolderService(BaseService):
    """Service for managing UiPath Folders.

    A folder represents a single area for data organization
    and access control - it is created when you need to categorize, manage, and enforce authorization rules for a group
    of UiPath resources (i.e. processes, assets, connections, storage buckets etc.) or other folders
    """

    def __init__(self, config: Config, execution_context: ExecutionContext) -> None:
        super().__init__(config=config, execution_context=execution_context)

    @traced(name="folder_retrieve_key_by_folder_path", run_type="uipath")
    @deprecated("Use retrieve_key instead")
    def retrieve_key_by_folder_path(self, folder_path: str) -> Optional[str]:
        return self.retrieve_key(folder_path=folder_path)

    @traced(name="folder_retrieve_key", run_type="uipath")
    def retrieve_key(self, *, folder_path: str) -> Optional[str]:
        """Retrieve the folder key by folder path with pagination support.

        Args:
            folder_path: The fully qualified folder path to search for.

        Returns:
            The folder key if found, None otherwise.
        """
        skip = 0
        take = 20

        while True:
            spec = self._retrieve_spec(folder_path, skip=skip, take=take)
            response = self.request(
                spec.method,
                url=spec.endpoint,
                params=spec.params,
            ).json()

            # Search for the folder in current page
            folder_key = next(
                (
                    item["Key"]
                    for item in response["PageItems"]
                    if item["FullyQualifiedName"] == folder_path
                ),
                None,
            )

            if folder_key is not None:
                return folder_key

            page_items = response["PageItems"]
            if len(page_items) < take:
                break

            skip += take

        return None

    def _retrieve_spec(
        self, folder_path: str, *, skip: int = 0, take: int = 20
    ) -> RequestSpec:
        folder_name = folder_path.split("/")[-1]
        return RequestSpec(
            method="GET",
            endpoint=Endpoint(
                "orchestrator_/api/FoldersNavigation/GetFoldersForCurrentUser"
            ),
            params={
                "searchText": folder_name,
                "skip": skip,
                "take": take,
            },
        )



import asyncio
import http.server
import json
import os
import socketserver
import threading
import time
from typing import Optional

from ._oidc_utils import OidcUtils

# Server port
PORT = 6234


# Custom exception for token received
class TokenReceivedSignal(Exception):
    """Exception raised when a token is successfully received."""

    def __init__(self, token_data):
        self.token_data = token_data
        super().__init__("Token received successfully")


def make_request_handler_class(state, code_verifier, token_callback, domain):
    class SimpleHTTPSRequestHandler(http.server.SimpleHTTPRequestHandler):
        """Simple HTTPS request handler that serves static files."""

        def log_message(self, format, *args) -> None:
            # do nothing
            pass

        def do_POST(self):
            """Handle POST requests to /set_token."""
            if self.path == "/set_token":
                content_length = int(self.headers["Content-Length"])
                post_data = self.rfile.read(content_length)
                token_data = json.loads(post_data.decode("utf-8"))

                self.send_response(200)
                self.end_headers()
                self.wfile.write(b"Token received")

                time.sleep(1)

                token_callback(token_data)
            elif self.path == "/log":
                content_length = int(self.headers["Content-Length"])
                post_data = self.rfile.read(content_length)
                logs = json.loads(post_data.decode("utf-8"))
                # Write logs to .uipath/.error_log file
                uipath_dir = os.path.join(os.getcwd(), ".uipath")
                os.makedirs(uipath_dir, exist_ok=True)
                error_log_path = os.path.join(uipath_dir, ".error_log")

                with open(error_log_path, "a", encoding="utf-8") as f:
                    f.write(
                        f"\n--- Authentication Error Log {time.strftime('%Y-%m-%d %H:%M:%S')} ---\n"
                    )
                    json.dump(logs, f, indent=2)
                    f.write("\n")
                self.send_response(200)
                self.end_headers()
                self.wfile.write(b"Log received")
            else:
                self.send_error(404, "Path not found")

        def do_GET(self):
            """Handle GET requests by serving index.html."""
            # Always serve index.html regardless of the path
            try:
                index_path = os.path.join(os.path.dirname(__file__), "index.html")
                with open(index_path, "r") as f:
                    content = f.read()

                # Get the redirect URI from auth config
                auth_config = OidcUtils.get_auth_config()
                redirect_uri = auth_config["redirect_uri"]

                content = content.replace("__PY_REPLACE_EXPECTED_STATE__", state)
                content = content.replace("__PY_REPLACE_CODE_VERIFIER__", code_verifier)
                content = content.replace("__PY_REPLACE_REDIRECT_URI__", redirect_uri)
                content = content.replace(
                    "__PY_REPLACE_CLIENT_ID__", auth_config["client_id"]
                )
                content = content.replace("__PY_REPLACE_DOMAIN__", domain)

                self.send_response(200)
                self.send_header("Content-Type", "text/html")
                self.send_header("Content-Length", str(len(content)))
                self.end_headers()
                self.wfile.write(content.encode("utf-8"))
            except FileNotFoundError:
                self.send_error(404, "File not found")

        def end_headers(self):
            self.send_header("Access-Control-Allow-Origin", "*")
            self.send_header("Access-Control-Allow-Methods", "GET, POST, OPTIONS")
            self.send_header("Access-Control-Allow-Headers", "Content-Type")
            super().end_headers()

        def do_OPTIONS(self):
            self.send_response(200)
            self.end_headers()

    return SimpleHTTPSRequestHandler


class HTTPServer:
    def __init__(self, port=6234):
        """Initialize HTTP server with configurable parameters.

        Args:
            port (int, optional): Port number to run the server on. Defaults to 6234.
        """
        self.current_path = os.path.dirname(os.path.abspath(__file__))
        self.port = port
        self.httpd: Optional[socketserver.TCPServer] = None
        self.token_data = None
        self.should_shutdown = False
        self.token_received_event: Optional[asyncio.Event] = None
        self.loop = None

    def token_received_callback(self, token_data):
        """Callback for when a token is received.

        Args:
            token_data (dict): The received token data.
        """
        self.token_data = token_data
        if self.token_received_event and self.loop:
            self.loop.call_soon_threadsafe(self.token_received_event.set)

    def create_server(self, state, code_verifier, domain):
        """Create and configure the HTTP server.

        Args:
            state (str): The OAuth state parameter.
            code_verifier (str): The PKCE code verifier.
            domain (str): The domain for authentication.

        Returns:
            socketserver.TCPServer: The configured HTTP server.
        """
        # Create server with address reuse
        socketserver.TCPServer.allow_reuse_address = True
        handler = make_request_handler_class(
            state, code_verifier, self.token_received_callback, domain
        )
        self.httpd = socketserver.TCPServer(("", self.port), handler)
        return self.httpd

    def _run_server(self):
        """Run server loop in thread."""
        try:
            while not self.should_shutdown and self.httpd:
                self.httpd.handle_request()
        except Exception:
            # Server might be closed, that's fine
            pass

    async def start(self, state, code_verifier, domain):
        """Start the server.

        Args:
            state (str): The OAuth state parameter.
            code_verifier (str): The PKCE code verifier.
            domain (str): The domain for authentication.

        Returns:
            dict: The received token data or an empty dict if no token was received.
        """
        if not self.httpd:
            self.create_server(state, code_verifier, domain)

        self.token_received_event = asyncio.Event()
        self.loop = asyncio.get_event_loop()

        # Run server in daemon thread
        server_thread = threading.Thread(target=self._run_server, daemon=True)
        server_thread.start()

        try:
            # Wait indefinitely for token received event or interrupt
            await self.token_received_event.wait()
        finally:
            self.stop()

        return self.token_data if self.token_data else {}

    def stop(self):
        """Stop the server gracefully and cleanup resources."""
        self.should_shutdown = True
        if self.httpd:
            self.httpd.server_close()
            self.httpd = None



from enum import IntEnum
from typing import Any, Dict, List

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


class EvaluationItem(BaseModel):
    """Individual evaluation item within an evaluation set."""

    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    id: str
    name: str
    inputs: Dict[str, Any]
    expected_output: Dict[str, Any]
    expected_agent_behavior: str = ""
    simulation_instructions: str = ""
    simulate_input: bool = False
    input_generation_instructions: str = ""
    simulate_tools: bool = False
    tools_to_simulate: List[str] = Field(default_factory=list)
    eval_set_id: str
    created_at: str
    updated_at: str


class EvaluationSet(BaseModel):
    """Complete evaluation set model."""

    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    id: str
    file_name: str
    evaluator_refs: List[str] = Field(default_factory=list)
    evaluations: List[EvaluationItem] = Field(default_factory=list)
    name: str
    batch_size: int = 10
    timeout_minutes: int = 20
    model_settings: List[Dict[str, Any]] = Field(default_factory=list)
    created_at: str
    updated_at: str

    def extract_selected_evals(self, eval_ids) -> None:
        selected_evals: list[EvaluationItem] = []
        for evaluation in self.evaluations:
            if evaluation.id in eval_ids:
                selected_evals.append(evaluation)
                eval_ids.remove(evaluation.id)
        if len(eval_ids) > 0:
            raise ValueError("Unknown evaluation ids: {}".format(eval_ids))
        self.evaluations = selected_evals


class EvaluationStatus(IntEnum):
    PENDING = 0
    IN_PROGRESS = 1
    COMPLETED = 2



# type: ignore
import click

from ..telemetry import track
from .cli_pack import pack
from .cli_publish import publish


@click.command()
@click.option(
    "--tenant",
    "-t",
    "feed",
    flag_value="tenant",
    help="Whether to publish to the tenant package feed",
)
@click.option(
    "--my-workspace",
    "-w",
    "feed",
    flag_value="personal",
    help="Whether to publish to the personal workspace",
)
@click.argument("root", type=str, default="./")
@track
def deploy(root, feed):
    """Pack and publish the project."""
    ctx = click.get_current_context()
    ctx.invoke(pack, root=root)
    ctx.invoke(publish, feed=feed)



import importlib
import inspect
import json
import logging
from contextvars import ContextVar
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple

from opentelemetry import context, trace
from opentelemetry.trace import set_span_in_context

from ._utils import _SpanUtils

logger = logging.getLogger(__name__)

_tracer_instance: Optional[trace.Tracer] = None
# ContextVar to track the currently active span for nesting
_active_traced_span: ContextVar[Optional[trace.Span]] = ContextVar(
    "_active_traced_span", default=None
)


def get_tracer() -> trace.Tracer:
    """Lazily initializes and returns the tracer instance."""
    global _tracer_instance
    if _tracer_instance is None:
        _tracer_instance = trace.get_tracer(__name__)
    return _tracer_instance


class TracingManager:
    """Static utility class to manage tracing implementations and decorated functions."""

    # Registry to track original functions, decorated functions, and their parameters
    # Each entry is (original_func, decorated_func, params)
    _traced_registry: List[Tuple[Callable[..., Any], Callable[..., Any], Any]] = []

    # Custom tracer implementation
    _custom_tracer_implementation = None  # Custom span provider function
    _current_span_provider: Optional[Callable[[], Any]] = None

    @classmethod
    def get_custom_tracer_implementation(cls):
        """Get the currently set custom tracer implementation."""
        return cls._custom_tracer_implementation

    @classmethod
    def register_current_span_provider(
        cls, current_span_provider: Optional[Callable[[], Any]]
    ):
        """Register a custom current span provider function.

        Args:
            current_span_provider: A function that returns the current span from an external
                                 tracing framework. If None, no custom span parenting will be used.
        """
        cls._current_span_provider = current_span_provider

    @staticmethod
    def get_parent_context():
        # Always use the currently active OTel span if valid (recursion / children)
        current_span = trace.get_current_span()
        if current_span is not None and current_span.get_span_context().is_valid:
            return set_span_in_context(current_span)

        # Only for the very top-level call, fallback to LangGraph span
        if TracingManager._current_span_provider is not None:
            try:
                external_span = TracingManager._current_span_provider()
                if external_span is not None:
                    return set_span_in_context(external_span)
            except Exception as e:
                logger.warning(f"Error getting current span from provider: {e}")

        # Last fallback
        return context.get_current()

    @classmethod
    def register_traced_function(cls, original_func, decorated_func, params):
        """Register a function decorated with @traced and its parameters.

        Args:
            original_func: The original function before decoration
            decorated_func: The function after decoration
            params: The parameters used for tracing
        """
        cls._traced_registry.append((original_func, decorated_func, params))

    @classmethod
    def reapply_traced_decorator(cls, tracer_implementation):
        """Reapply a different tracer implementation to all functions previously decorated with @traced.

        Args:
            tracer_implementation: A function that takes the same parameters as _opentelemetry_traced
                                 and returns a decorator. If None, reverts to default implementation.
        """
        tracer_implementation = tracer_implementation or _opentelemetry_traced
        cls._custom_tracer_implementation = tracer_implementation

        # Work with a copy of the registry to avoid modifying it during iteration
        registry_copy = cls._traced_registry.copy()

        for original_func, decorated_func, params in registry_copy:
            # Apply the new decorator with the same parameters
            supported_params = _get_supported_params(tracer_implementation, params)
            new_decorated_func = tracer_implementation(**supported_params)(
                original_func
            )

            logger.debug(
                f"Reapplying decorator to {original_func.__name__}, from {decorated_func.__name__}"
            )

            # If this is a method on a class, we need to update the class
            if hasattr(original_func, "__self__") and hasattr(
                original_func, "__func__"
            ):
                setattr(
                    original_func.__self__.__class__,
                    original_func.__name__,
                    new_decorated_func.__get__(
                        original_func.__self__, original_func.__self__.__class__
                    ),
                )
            else:
                # Replace the function in its module
                if hasattr(original_func, "__module__") and hasattr(
                    original_func, "__qualname__"
                ):
                    try:
                        module = importlib.import_module(original_func.__module__)
                        parts = original_func.__qualname__.split(".")

                        # Handle nested objects
                        obj = module
                        for part in parts[:-1]:
                            obj = getattr(obj, part)

                        setattr(obj, parts[-1], new_decorated_func)

                        # Update the registry entry for this function
                        # Find the index and replace with updated entry
                        for i, (orig, _dec, _p) in enumerate(cls._traced_registry):
                            if orig is original_func:
                                cls._traced_registry[i] = (
                                    original_func,
                                    new_decorated_func,
                                    params,
                                )
                                break
                    except (ImportError, AttributeError) as e:
                        # Log the error but continue processing other functions
                        logger.warning(f"Error reapplying decorator: {e}")
                        continue


def _default_input_processor(inputs):
    """Default input processor that doesn't log any actual input data."""
    return {"redacted": "Input data not logged for privacy/security"}


def _default_output_processor(outputs):
    """Default output processor that doesn't log any actual output data."""
    return {"redacted": "Output data not logged for privacy/security"}


def wait_for_tracers():
    """Wait for all tracers to finish."""
    trace.get_tracer_provider().shutdown()  # type: ignore


def _opentelemetry_traced(
    name: Optional[str] = None,
    run_type: Optional[str] = None,
    span_type: Optional[str] = None,
    input_processor: Optional[Callable[..., Any]] = None,
    output_processor: Optional[Callable[..., Any]] = None,
):
    """Default tracer implementation using OpenTelemetry."""

    def decorator(func):
        trace_name = name or func.__name__

        def get_parent_context():
            """Return a context object for starting the new span."""
            current_span = _active_traced_span.get()
            if current_span is not None and current_span.get_span_context().is_valid:
                return set_span_in_context(current_span)

            if TracingManager._current_span_provider is not None:
                try:
                    external_span = TracingManager._current_span_provider()
                    if external_span is not None:
                        return set_span_in_context(external_span)
                except Exception as e:
                    logger.warning(f"Error getting current span from provider: {e}")

            return context.get_current()

        # --------- Sync wrapper ---------
        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            ctx = get_parent_context()
            span_cm = get_tracer().start_as_current_span(trace_name, context=ctx)
            span = span_cm.__enter__()
            token = _active_traced_span.set(span)
            try:
                span.set_attribute("span_type", span_type or "function_call_sync")
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                if input_processor:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)

                result = func(*args, **kwargs)
                output = output_processor(result) if output_processor else result
                span.set_attribute("output", json.dumps(output, default=str))
                return result
            except Exception as e:
                span.record_exception(e)
                span.set_status(
                    trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                )
                raise
            finally:
                _active_traced_span.reset(token)
                span_cm.__exit__(None, None, None)

        # --------- Async wrapper ---------
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            ctx = get_parent_context()
            span_cm = get_tracer().start_as_current_span(trace_name, context=ctx)
            span = span_cm.__enter__()
            token = _active_traced_span.set(span)
            try:
                span.set_attribute("span_type", span_type or "function_call_async")
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                if input_processor:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)

                result = await func(*args, **kwargs)
                output = output_processor(result) if output_processor else result
                span.set_attribute("output", json.dumps(output, default=str))
                return result
            except Exception as e:
                span.record_exception(e)
                span.set_status(
                    trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                )
                raise
            finally:
                _active_traced_span.reset(token)
                span_cm.__exit__(None, None, None)

        # --------- Generator wrapper ---------
        @wraps(func)
        def generator_wrapper(*args, **kwargs):
            ctx = get_parent_context()
            span_cm = get_tracer().start_as_current_span(trace_name, context=ctx)
            span = span_cm.__enter__()
            token = _active_traced_span.set(span)
            try:
                span.set_attribute(
                    "span_type", span_type or "function_call_generator_sync"
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                if input_processor:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)

                outputs = []
                for item in func(*args, **kwargs):
                    outputs.append(item)
                    span.add_event(f"Yielded: {item}")
                    yield item
                output = output_processor(outputs) if output_processor else outputs
                span.set_attribute("output", json.dumps(output, default=str))
            except Exception as e:
                span.record_exception(e)
                span.set_status(
                    trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                )
                raise
            finally:
                _active_traced_span.reset(token)
                span_cm.__exit__(None, None, None)

        # --------- Async generator wrapper ---------
        @wraps(func)
        async def async_generator_wrapper(*args, **kwargs):
            ctx = get_parent_context()
            span_cm = get_tracer().start_as_current_span(trace_name, context=ctx)
            span = span_cm.__enter__()
            token = _active_traced_span.set(span)
            try:
                span.set_attribute(
                    "span_type", span_type or "function_call_generator_async"
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                if input_processor:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)

                outputs = []
                async for item in func(*args, **kwargs):
                    outputs.append(item)
                    span.add_event(f"Yielded: {item}")
                    yield item
                output = output_processor(outputs) if output_processor else outputs
                span.set_attribute("output", json.dumps(output, default=str))
            except Exception as e:
                span.record_exception(e)
                span.set_status(
                    trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                )
                raise
            finally:
                _active_traced_span.reset(token)
                span_cm.__exit__(None, None, None)

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        elif inspect.isgeneratorfunction(func):
            return generator_wrapper
        elif inspect.isasyncgenfunction(func):
            return async_generator_wrapper
        else:
            return sync_wrapper

    return decorator


def _get_supported_params(tracer_impl, params):
    """Extract the parameters supported by the tracer implementation.

    Args:
        tracer_impl: The tracer implementation function or callable
        params: Dictionary of parameters to check

    Returns:
        Dictionary containing only parameters supported by the tracer implementation
    """
    supported_params = {}
    if hasattr(tracer_impl, "__code__"):
        # For regular functions
        impl_signature = inspect.signature(tracer_impl)
        for param_name, param_value in params.items():
            if param_name in impl_signature.parameters and param_value is not None:
                supported_params[param_name] = param_value
    elif callable(tracer_impl):
        # For callable objects
        impl_signature = inspect.signature(tracer_impl.__call__)
        for param_name, param_value in params.items():
            if param_name in impl_signature.parameters and param_value is not None:
                supported_params[param_name] = param_value
    else:
        # If we can't inspect, pass all parameters and let the function handle it
        supported_params = params

    return supported_params


def traced(
    name: Optional[str] = None,
    run_type: Optional[str] = None,
    span_type: Optional[str] = None,
    input_processor: Optional[Callable[..., Any]] = None,
    output_processor: Optional[Callable[..., Any]] = None,
    hide_input: bool = False,
    hide_output: bool = False,
):
    """Decorator that will trace function invocations.

    Args:
        run_type: Optional string to categorize the run type
        span_type: Optional string to categorize the span type
        input_processor: Optional function to process function inputs before recording
            Should accept a dictionary of inputs and return a processed dictionary
        output_processor: Optional function to process function outputs before recording
            Should accept the function output and return a processed value
        hide_input: If True, don't log any input data
        hide_output: If True, don't log any output data
    """
    # Apply default processors selectively based on hide flags
    if hide_input:
        input_processor = _default_input_processor
    if hide_output:
        output_processor = _default_output_processor

    # Store the parameters for later reapplication
    params = {
        "name": name,
        "run_type": run_type,
        "span_type": span_type,
        "input_processor": input_processor,
        "output_processor": output_processor,
    }

    # Check for custom implementation first
    custom_implementation = TracingManager.get_custom_tracer_implementation()
    tracer_impl: Any = (
        custom_implementation if custom_implementation else _opentelemetry_traced
    )

    def decorator(func):
        # Check which parameters are supported by the tracer_impl
        supported_params = _get_supported_params(tracer_impl, params)

        # Decorate the function with only supported parameters
        decorated_func = tracer_impl(**supported_params)(func)

        # Register both original and decorated function with parameters
        TracingManager.register_traced_function(func, decorated_func, params)
        return decorated_func

    return decorator



"""Models for evaluation framework including execution data and evaluation results."""

from enum import IntEnum
from typing import Annotated, Any, Dict, Literal, Optional, Union

from opentelemetry.sdk.trace import ReadableSpan
from pydantic import BaseModel, ConfigDict, Field


class AgentExecution(BaseModel):
    """Represents the execution data of an agent for evaluation purposes."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    agent_input: Optional[Dict[str, Any]]
    agent_output: Dict[str, Any]
    agent_trace: list[ReadableSpan]


class LLMResponse(BaseModel):
    """Response from an LLM evaluator."""

    score: float
    justification: str


class ScoreType(IntEnum):
    """Types of evaluation scores."""

    BOOLEAN = 0
    NUMERICAL = 1
    ERROR = 2


class BaseEvaluationResult(BaseModel):
    """Base class for evaluation results."""

    details: Optional[str] = None
    # this is marked as optional, as it is populated inside the 'measure_execution_time' decorator
    evaluation_time: Optional[float] = None


class BooleanEvaluationResult(BaseEvaluationResult):
    """Result of a boolean evaluation."""

    score: bool
    score_type: Literal[ScoreType.BOOLEAN] = ScoreType.BOOLEAN


class NumericEvaluationResult(BaseEvaluationResult):
    """Result of a numerical evaluation."""

    score: float
    score_type: Literal[ScoreType.NUMERICAL] = ScoreType.NUMERICAL


class ErrorEvaluationResult(BaseEvaluationResult):
    """Result of an error evaluation."""

    score: float = 0.0
    score_type: Literal[ScoreType.ERROR] = ScoreType.ERROR


EvaluationResult = Annotated[
    Union[BooleanEvaluationResult, NumericEvaluationResult, ErrorEvaluationResult],
    Field(discriminator="score_type"),
]


class EvalItemResult(BaseModel):
    """Result of a single evaluation item."""

    evaluator_name: str
    result: EvaluationResult


class EvaluatorCategory(IntEnum):
    """Types of evaluators."""

    Deterministic = 0
    LlmAsAJudge = 1
    AgentScorer = 2
    Trajectory = 3

    @classmethod
    def from_int(cls, value):
        """Construct EvaluatorCategory from an int value."""
        if value in cls._value2member_map_:
            return cls(value)
        else:
            raise ValueError(f"{value} is not a valid EvaluatorCategory value")


class EvaluatorType(IntEnum):
    """Subtypes of evaluators."""

    Unknown = 0
    Equals = 1
    Contains = 2
    Regex = 3
    Factuality = 4
    Custom = 5
    JsonSimilarity = 6
    Trajectory = 7
    ContextPrecision = 8
    Faithfulness = 9

    @classmethod
    def from_int(cls, value):
        """Construct EvaluatorCategory from an int value."""
        if value in cls._value2member_map_:
            return cls(value)
        else:
            raise ValueError(f"{value} is not a valid EvaluatorType value")



"""Message-level events."""

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from .content import UiPathConversationContentPart, UiPathConversationContentPartEvent
from .tool import UiPathConversationToolCall, UiPathConversationToolCallEvent


class UiPathConversationMessageStartEvent(BaseModel):
    """Signals the start of a message within an exchange."""

    exchange_sequence: Optional[int] = Field(None, alias="exchangeSequence")
    timestamp: Optional[str] = None
    role: str
    meta_data: Optional[Dict[str, Any]] = Field(None, alias="metaData")

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


class UiPathConversationMessageEndEvent(BaseModel):
    """Signals the end of a message."""

    meta_data: Optional[Dict[str, Any]] = Field(None, alias="metaData")

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


class UiPathConversationMessageEvent(BaseModel):
    """Encapsulates sub-events related to a message."""

    message_id: str = Field(..., alias="messageId")
    start: Optional[UiPathConversationMessageStartEvent] = None
    end: Optional[UiPathConversationMessageEndEvent] = None
    content_part: Optional[UiPathConversationContentPartEvent] = Field(
        None, alias="contentPart"
    )
    tool_call: Optional[UiPathConversationToolCallEvent] = Field(None, alias="toolCall")
    meta_event: Optional[Dict[str, Any]] = Field(None, alias="metaEvent")

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)


class UiPathConversationMessage(BaseModel):
    """Represents a single message within an exchange."""

    message_id: str = Field(..., alias="messageId")
    role: str
    content_parts: Optional[List[UiPathConversationContentPart]] = Field(
        None, alias="contentParts"
    )
    tool_calls: Optional[List[UiPathConversationToolCall]] = Field(
        None, alias="toolCalls"
    )
    created_at: str = Field(..., alias="createdAt")
    updated_at: str = Field(..., alias="updatedAt")

    model_config = ConfigDict(validate_by_name=True, validate_by_alias=True)



"""Trajectory evaluator for analyzing execution paths and decision sequences."""

from typing import TypeVar

from uipath.eval.models import EvaluationResult

from ..models.models import AgentExecution
from .base_evaluator import BaseEvaluator

T = TypeVar("T")


class TrajectoryEvaluator(BaseEvaluator[T]):
    """Evaluator that analyzes the trajectory/path taken to reach outputs."""

    async def evaluate(
        self, agent_execution: AgentExecution, evaluation_criteria: T
    ) -> EvaluationResult:
        """Evaluate using trajectory analysis.

        Analyzes the execution path and decision sequence taken by the agent
        to assess the quality of the reasoning process.

        Args:
            agent_execution: The execution details containing:
                - agent_input: The input received by the agent
                - actual_output: The actual output from the agent
                - spans: The execution spans to use for the evaluation
            evaluation_criteria: The criteria to evaluate
        Returns:
            EvaluationResult: Score based on trajectory analysis

        Raises:
            NotImplementedError: This evaluator is not yet implemented
        """
        raise NotImplementedError()



import json
import logging
from typing import Any, Dict

from .._config import Config
from .._execution_context import ExecutionContext
from .._utils import Endpoint, RequestSpec, infer_bindings
from ..models import Connection, ConnectionToken, EventArguments
from ..tracing._traced import traced
from ._base_service import BaseService

logger: logging.Logger = logging.getLogger("uipath")


class ConnectionsService(BaseService):
    """Service for managing UiPath external service connections.

    This service provides methods to retrieve direct connection information retrieval
    and secure token management.
    """

    def __init__(self, config: Config, execution_context: ExecutionContext) -> None:
        super().__init__(config=config, execution_context=execution_context)

    @traced(
        name="connections_retrieve",
        run_type="uipath",
        hide_output=True,
    )
    def retrieve(self, key: str) -> Connection:
        """Retrieve connection details by its key.

        This method fetches the configuration and metadata for a connection,
        which can be used to establish communication with an external service.

        Args:
            key (str): The unique identifier of the connection to retrieve.

        Returns:
            Connection: The connection details, including configuration parameters
                and authentication information.
        """
        spec = self._retrieve_spec(key)
        response = self.request(spec.method, url=spec.endpoint)
        return Connection.model_validate(response.json())

    @traced(
        name="connections_retrieve",
        run_type="uipath",
        hide_output=True,
    )
    async def retrieve_async(self, key: str) -> Connection:
        """Asynchronously retrieve connection details by its key.

        This method fetches the configuration and metadata for a connection,
        which can be used to establish communication with an external service.

        Args:
            key (str): The unique identifier of the connection to retrieve.

        Returns:
            Connection: The connection details, including configuration parameters
                and authentication information.
        """
        spec = self._retrieve_spec(key)
        response = await self.request_async(spec.method, url=spec.endpoint)
        return Connection.model_validate(response.json())

    @traced(
        name="connections_retrieve_token",
        run_type="uipath",
        hide_output=True,
    )
    def retrieve_token(self, key: str) -> ConnectionToken:
        """Retrieve an authentication token for a connection.

        This method obtains a fresh authentication token that can be used to
        communicate with the external service. This is particularly useful for
        services that use token-based authentication.

        Args:
            key (str): The unique identifier of the connection.

        Returns:
            ConnectionToken: The authentication token details, including the token
                value and any associated metadata.
        """
        spec = self._retrieve_token_spec(key)
        response = self.request(spec.method, url=spec.endpoint, params=spec.params)
        return ConnectionToken.model_validate(response.json())

    @traced(
        name="connections_retrieve_token",
        run_type="uipath",
        hide_output=True,
    )
    async def retrieve_token_async(self, key: str) -> ConnectionToken:
        """Asynchronously retrieve an authentication token for a connection.

        This method obtains a fresh authentication token that can be used to
        communicate with the external service. This is particularly useful for
        services that use token-based authentication.

        Args:
            key (str): The unique identifier of the connection.

        Returns:
            ConnectionToken: The authentication token details, including the token
                value and any associated metadata.
        """
        spec = self._retrieve_token_spec(key)
        response = await self.request_async(
            spec.method, url=spec.endpoint, params=spec.params
        )
        return ConnectionToken.model_validate(response.json())

    @traced(
        name="connections_retrieve_event_payload",
        run_type="uipath",
    )
    @infer_bindings(resource_type="ignored", ignore=True)
    def retrieve_event_payload(self, event_args: EventArguments) -> Dict[str, Any]:
        """Retrieve event payload from UiPath Integration Service.

        Args:
            event_args (EventArguments): The event arguments. Should be passed along from the job's input.

        Returns:
            Dict[str, Any]: The event payload data
        """
        if not event_args.additional_event_data:
            raise ValueError("additional_event_data is required")

        # Parse additional event data to get event id
        event_data = json.loads(event_args.additional_event_data)

        event_id = None
        if "processedEventId" in event_data:
            event_id = event_data["processedEventId"]
        elif "rawEventId" in event_data:
            event_id = event_data["rawEventId"]
        else:
            raise ValueError("Event Id not found in additional event data")

        # Build request URL using connection token's API base URI
        spec = self._retrieve_event_payload_spec("v1", event_id)

        response = self.request(spec.method, url=spec.endpoint)

        return response.json()

    @traced(
        name="connections_retrieve_event_payload",
        run_type="uipath",
    )
    @infer_bindings(resource_type="ignored", ignore=True)
    async def retrieve_event_payload_async(
        self, event_args: EventArguments
    ) -> Dict[str, Any]:
        """Retrieve event payload from UiPath Integration Service.

        Args:
            event_args (EventArguments): The event arguments. Should be passed along from the job's input.

        Returns:
            Dict[str, Any]: The event payload data
        """
        if not event_args.additional_event_data:
            raise ValueError("additional_event_data is required")

        # Parse additional event data to get event id
        event_data = json.loads(event_args.additional_event_data)

        event_id = None
        if "processedEventId" in event_data:
            event_id = event_data["processedEventId"]
        elif "rawEventId" in event_data:
            event_id = event_data["rawEventId"]
        else:
            raise ValueError("Event Id not found in additional event data")

        # Build request URL using connection token's API base URI
        spec = self._retrieve_event_payload_spec("v1", event_id)

        response = await self.request_async(spec.method, url=spec.endpoint)

        return response.json()

    def _retrieve_event_payload_spec(self, version: str, event_id: str) -> RequestSpec:
        return RequestSpec(
            method="GET",
            endpoint=Endpoint(f"/elements_/{version}/events/{event_id}"),
        )

    def _retrieve_spec(self, key: str) -> RequestSpec:
        return RequestSpec(
            method="GET",
            endpoint=Endpoint(f"/connections_/api/v1/Connections/{key}"),
        )

    def _retrieve_token_spec(self, key: str) -> RequestSpec:
        return RequestSpec(
            method="GET",
            endpoint=Endpoint(f"/connections_/api/v1/Connections/{key}/token"),
            params={"tokenType": "direct"},
        )
